﻿--------------------------------------------------------------------------------------------------------
-- Moncaí's Quest Tracker
-- Author: Moncaí
-- Copyright (C) 2006-2009 moncai AT mgtx DOT net, all rights reserved.
-- Date: $Date: 2009-08-18 14:56:08 +0000 (Tue, 18 Aug 2009) $, $Rev: 18 $
--------------------------------------------------------------------------------------------------------
local MQT = LibStub("AceAddon-3.0"):NewAddon("MoncaiQuestTracker", "AceConsole-3.0");
_G["MQT"] = MQT;
local L = LibStub("AceLocale-3.0"):GetLocale("MoncaiQuestTracker", false);
--------------------------------------------------------------------------------------------------------
--
-- constants
-- 
--------------------------------------------------------------------------------------------------------

-- This is specific code that doesn't belong in the core at all...
MQT.WATCHFRAME_MINIMUMWIDTH = 220;
MQT.WATCHFRAME_QUEST_OFFSET = 5;
MQT.WATCHFRAME_QUEST_WITH_ITEM_HEIGHT = WATCHFRAME_LINEHEIGHT * 2;
-- si sihT

MQT.MAX_PARTY = 5

MQT.LEVEL_FORMAT = L["[%d] %s"];
MQT.MSG_C = { r = 1, g = 1, b = 0.8 }

MQT.MSG_PREFIX = "MQTm";
MQT.QST_PREFIX = "MQTq";
MQT.CMD_PREFIX = "MQTu";
MQT.CMD_UPDATE = "!";
MQT.CMD_BULK = "b";
MQT.CMD_PULL = "^";

MQT.MATCH3 = L["(.*): (%d+) ?/ ?(%d+)"]
MQT.MATCH1 = L[".*: (%d+) ?/ ?%d+"]
MQT.MATCHa = L["(.*): %d+ ?/ ?%d+"]

--------------------------------------------------------------------------------------------------------
--
-- Library Function Imports
--
--------------------------------------------------------------------------------------------------------
local ML = LibStub("MoncaiLib");
local U = ML.util;

local RGBPercToHex = U.RGBPercToHex;
local RGBToHex = U.RGBToHex;
local wrap = U.wrap;
local unwrap = U.unwrap;

local ChatThrottleLib = ChatThrottleLib;

--------------------------------------------------------------------------------------------------------
--
-- VARIABLES
--
--------------------------------------------------------------------------------------------------------

-- Saved variable
MQTC = {
	QList = { },
	ColorTitle = 1,
	ColorStatus = 0,
	AutoAdd = 1,
	Enabled = 1,
	ShowLevel = 1,
	RemoveComplete = 0,
	OnHold = 0,
	LineMax = 30,
	Communicate = 1,
};

-- Working memory variable
MQT.workspace = {
	watches = -1,
	MaxLines = 35,
	Comms = {},
	hash = {},
};
local MQTW = MQT.workspace;

--------------------------------------------------------------------------------------------------------
--
-- Buddylist
--
--------------------------------------------------------------------------------------------------------

local BuddyList = {
	colors = {
		[1] = { r = 111, g = 111, b = 255 },
		[2] = { r = 111, g = 255, b = 111 },
		[3] = { r = 255, g = 111, b = 111 },
		[4] = { r = 255, g = 255, b = 111 },
		fmt = function(self, buddynr, nr)
			return string.format(" |cff%02x%02x%02x%d|r", self[buddynr].r, self[buddynr].g, self[buddynr].b, nr)
		end,
	},
	buddies = {},
	questlb = {},
	
	cleanup = function(self, bnr)
		-- remove all possibly old references to this buddy
		for title, quest in pairs(self.questlb) do
			for idx, lb in ipairs(quest) do
				lb[bnr] = nil
			end
		end
	end,
	
	removeBuddy = function(self, name)
		local i = self.buddies[name] or 0
		if i == 0 then return end
		self.buddies[name] = nil
		self.buddies[i] = nil
		
		-- remove all references to this buddy
		self:cleanup(i)

		if MQT.db.profile.verbosebuddy then Print(L["Goodbye, "] .. name) end
	end,

	addBuddy = function(self, name)
		for i = 1, MQT.MAX_PARTY, 1 do
			if not self.buddies[i] then
				self.buddies[i] = { ["name"] = name, questlog = {}, buffer = {}, watches = {} }
				self.buddies[name] = i
				if MQT.db.profile.verbosebuddy then Print(L["Welcome to the group, "] .. name .. "!") end
				return i
			end
		end
		if MQT.db.profile.verbosebuddy then Print(L["Could not add buddy, buffer full"]) end
		return false
	end,

	insertQuestPart = function(qtable, buddy, questTitle, level, lbidx, lbtext)	
		local quest = qtable[questTitle] or {}
		quest.level = level
		local lb = quest[lbidx] or {}
		lb.leaderboard = lbtext
		lb.nr = string.gmatch(lbtext, MQT.MATCH1)() or 0
		
		quest[lbidx] = lb
		qtable[questTitle] = quest
	end,
	
	receiveQuestPart = function(self, buddy, questTitle, level, lbidx, lbtext)
		local bnr = self.buddies[buddy] or self:addBuddy(buddy)
		if not bnr then
			if MQT.db.profile.verbosebuddy then Print(L["Could not add buddy, buffer full"]) end
			return
		end
		lbidx = 0 + lbidx
		
		self.insertQuestPart(self.buddies[bnr].buffer, buddy, questTitle, level, lbidx, lbtext)
		
	end,

	updateBuddyWatch = function (self, buddynr, questTitle, lbidx, nr)
		-- Store into easy to access thingy
		local q = self.questlb[questTitle] or {}
		local lb = q[lbidx] or {}
		lb[buddynr] = self.colors:fmt(buddynr, nr)
		
		q[lbidx] = lb
		self.questlb[questTitle] = q
	end,
	
	getBuddyWatch = function(self, questTitle, lbidx)
		local q = self.questlb[questTitle] or {}
		local lb = q[lbidx] or {}
		local str = "";
		for i = 1, MQT.MAX_PARTY, 1 do
			if lb[i] then
				str = str .. lb[i]
			end
		end

		return str
	end,
	
	updateQuestPart = function(self, buddy, questTitle, level, lbidx, lbtext)
		local bnr = self.buddies[buddy] or self:addBuddy(buddy)
		if not bnr then
			if MQT.db.profile.verbosebuddy then Print(L["Could not add buddy, buffer full"]) end
			return
		end
		lbidx = 0 + lbidx

		self.insertQuestPart(self.buddies[bnr].buffer, buddy, questTitle, level, lbidx, lbtext)
		self.insertQuestPart(self.buddies[bnr].questlog, buddy, questTitle, level, lbidx, lbtext)
		self:updateBuddyWatch( bnr, questTitle, lbidx, string.gmatch(lbtext, MQT.MATCH1)() )
	end,
	
	propagate = function(self, buddy)
		local bnr = self.buddies[buddy] or self:addBuddy(buddy)
		if not bnr then return end
		
		self.buddies[bnr].questlog = self.buddies[bnr].buffer
		self.buddies[bnr].buffer = {}
		
		
		self:cleanup(bnr)
		
		-- insert nice new ones
		for title, quest in pairs(self.buddies[bnr].questlog) do
			for idx, lb in ipairs(quest) do
				self:updateBuddyWatch(bnr, title, idx, lb.nr or 0)
			end
		end
	end,
	
	buddyCheck = function(self)
		local name
		local list = {}
		
		--Print("BuddyCheck!");
		-- get a list of all buddies
		for i = 1, MQT.MAX_PARTY, 1 do
			local name = UnitName("party" .. i)
			if name then
				list[name] = i
			end
		end

		-- remove buddies that are no longer around
		for i = 1, MQT.MAX_PARTY, 1 do
			if self.buddies[i] and not list[self.buddies[i].name] then
				self:removeBuddy(self.buddies[i].name)
			end
		end
		
		-- add new buddies
		for k,v in pairs(list) do
			if not self.buddies[k] then
				self:addBuddy(k)
			end
		end
	end,
}
MQT.BuddyList = BuddyList

-- Sample = {
-- 	["Person1"] = {
--		["Zalazane"] = {
--			level = 9,
--			[1] = {
--				leaderboard = "Zalazane's head: 0/1",
--				nr = 0,
--			}
--		}
--	}
--	["Person2"] = { },
--	["Person3"] = nil,
-- }

function MQT.Communicate() 
	local i=0, packet, msg;
	while (GetQuestLogTitle(i+1) ~= nil) do
		i = i + 1;
		local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(i);
		if (not header) then
			for j = 1, GetNumQuestLeaderBoards(i), 1 do
				msg = GetQuestLogLeaderBoard(j,i)
				packet = wrap(questTitle, level, j, msg)
				ChatThrottleLib:SendAddonMessage("BULK", MQT.QST_PREFIX, packet, "RAID");				
			end
		end
	end
	ChatThrottleLib:SendAddonMessage("BULK", MQT.QST_PREFIX, wrap(0,0,0,0, MQT.CMD_UPDATE), "RAID");
end

local FREQ = 60
local SPDFREQ = 5;
local tick = 55
local speedup = false;

-- Speed up the notification
function MQT.Notify()
	speedup = true;
end

function MQT.Update(self, elapsed)
	-- if QuestHelper then MQT.QuestHelperUpdateTimer(elapsed); end;

	tick = tick + elapsed;
	if (tick >= FREQ or (speedup and tick >= SPDFREQ)) then
		--Print("Comm channel open...");
		tick = 0;
		speedup = false;
		if (GetNumPartyMembers() > 0) then
			MQT.Communicate();
		end
	end
end

function MQT.Event(this, event, arg1, arg2, arg3, arg4, ...)
	local player = UnitName("player");
	local message = arg1;
	
	if ( event == "UI_INFO_MESSAGE" ) then
		local questUpdateText = gsub(message, "(.*): %d+ ?/ ?%d+","%1", 1);

		if ( questUpdateText ~= message) then
		
			--Print("UI Message looks like a quest objective");
			
			local questTitle, level, idx, qid, nrentries, lbtext, lb;
			questTitle = nil
			nrentries = GetNumQuestLogEntries()
			lbtext = string.gmatch(message, MQT.MATCHa)();
			
			for i=1, nrentries, 1 do
				for j = 1, GetNumQuestLeaderBoards(i), 1 do
					lb = string.gmatch( GetQuestLogLeaderBoard(j,i), MQT.MATCHa)()
					--Print("~" .. (lbtext or "") .."~?=~" .. (lb or "") .. "~")
					if (lbtext == lb) then
						questTitle, level = GetQuestLogTitle(i);
						idx = j;
						qid = i;
						break
					end
				end
				if (questTitle) then break end
			end
			-- if we have title, ship it off :)
			if (questTitle) then
				packet = wrap(questTitle, level, idx, message, MQT.CMD_UPDATE)

				-- go send
				ChatThrottleLib:SendAddonMessage("ALERT", MQT.QST_PREFIX, packet, "RAID");
				-- ChatThrottleLib:SendAddonMessage("NORMAL", MQT_CMD_PREFIX, MQT_CMD_UPDATE, "RAID");

				-- autotrack
				--Print(string.format("Adding %d: %s", idx, questTitle));
				
				AddQuestWatch(qid, true, true);
				QuestLog_Update();
				WatchFrame_Update();
				
			else
				if MQT.db.profile.debug then Print("Title not found") end
			end
		
			local msg, n, m = string.gmatch(message, "(.*): ?(%d+) ?/ ?(%d+)")();

			if msg and n and m then
				local str;
				if m-n > 0 then
					str = string.format(L["needs %1$d more %2$s "], m-n, msg);
				else
					str = string.format(L["is finished with %s "], msg);
				end
				ChatThrottleLib:SendAddonMessage("ALERT", MQT.MSG_PREFIX, str, "RAID");
				str = string.gsub(str, L["needs"], L["need"]);
				str = string.gsub(str, L["is finished"], L["are finished"]);
				if MQT.db.profile.verbosechat then Print(string.format("%s %s",L["You"],str)) end
			end
			
		else
			local messag = string.gsub(message, "Disco", "disco");
			ChatThrottleLib:SendAddonMessage("ALERT", MQT.MSG_PREFIX, L["has "] .. messag, "RAID");
			Print(message);
		end

	elseif ( event == "QUEST_LOG_UPDATE" or event == "UPDATE_FACTION" or event == "UNIT_QUEST_LOG_CHANGED") then
		-- currently not registered
		MQT.Notify();
		
	elseif ( event == "VARIABLES_LOADED" ) then
		-- show splash banner here?
		
	elseif ( event == "PARTY_MEMBERS_CHANGED" ) then
		BuddyList:buddyCheck()
		MQT.Notify()
		
	elseif ( event == "PLAYER_ENTERING_WORLD" ) then
		BuddyList:buddyCheck()
		if GetNumPartyMembers() > 1 then
			if MQT.db.profile.verbosebuddy then Print(L["Let's Party!"]) end
			MQT.Notify()
		end
		
	elseif ( event == "CHAT_MSG_ADDON" and arg4 ~= player ) then
		if arg1 == MQT.MSG_PREFIX then
			if MQT.db.profile.verbosechat then Print(string.format("%s %s", arg4, arg2), MQT.MSG_C.r, MQT.MSG_C.g, MQT.MSG_C.b) end
		
		elseif arg1 == MQT.QST_PREFIX then
			-- receive quest message
			
			local title, level, idx, str, cmd = unwrap(arg2);
			-- convert idx to number
			idx = 0 + idx;

			if cmd == MQT.CMD_UPDATE then
				--Print("Piggyback Update received from " .. arg4)
				
				if (title ~= "0") then
					BuddyList:updateQuestPart(arg4, title, level, idx, str)
				--else
				--	Print("Gots a naught packet m'lord, from " .. arg4)
				else
					BuddyList:propagate(arg4)
				end
				WatchFrame_Update()
			else
				BuddyList:receiveQuestPart(arg4, title, level, idx, str)
			end
			
		elseif arg1 == MQT.CMD_PREFIX then
			if arg2 == MQT.CMD_UPDATE then
				-- update aanroepen hier
				--Print("Update received from " .. arg4);
				BuddyList:propagate(arg4)
				WatchFrame_Update()
			else
				if MQT.db.profile.debug then Print(L["Command not found"]) end
			end
		end
	end	
	
end

function MQT.OnInitialize(self)
	local f,w;
	self.db = { profiles = {} } -- prevent nil errors
	
	--self:SetEnabledState(false);
	
	-- Initialize my library
	ML:init();

	Print(L["Moncaí's Quest Tracker loading, stand by..."]);

	if (self.Event == nil) then Print(L["PANIC!"]) end
	
	f = CreateFrame("Frame", "MQTEvent");
	f:SetScript("OnEvent", function(...) MQT.Event(...) end);
	f:SetScript("OnUpdate", function(...) MQT.Update(...) end);
	f:RegisterEvent("UI_INFO_MESSAGE");
	f:RegisterEvent("CHAT_MSG_ADDON");
	f:RegisterEvent("PARTY_MEMBERS_CHANGED");
	f:RegisterEvent("PLAYER_ENTERING_WORLD");
	
	MQTW.frame = f;
	
	
	local hooks = ML.api.hooking:new(self);
	self.hooks = hooks;

	hooks:HookGlobal("WATCHFRAME_MINIMUMWIDTH");
	hooks:HookGlobal("WATCHFRAME_QUEST_OFFSET");
	hooks:HookGlobal("WATCHFRAME_QUEST_WITH_ITEM_HEIGHT");
	hooks:HookGlobals(self.builtins);
	
	--[[ Stuff moved out to Configure.lua
	f = CreateFrame("Frame", "MQTMainOptionsFrame", UIParent);
	f.name = L["Moncaí Questtracker"];
	f.default = self.defaults;
	InterfaceOptions_AddCategory(f);
	self.optionspanel = f;
	
	LibStub("LibAboutPanel").new(L["Moncaí Questtracker"], "MoncaiQuestTracker");	
	self:RegisterChatCommand("rl", function() ReloadUI(); end, true);
	self:RegisterChatCommand("mqt", function() InterfaceOptionsFrame_OpenToCategory(self.optionspanel); end, true);
	self:RegisterChatCommand("mqtt", function() self:Toggle(); end, true);]]
end

-- revert to defaults here
function MQT.defaults()

end;

function MQT.OnToggle(self)
	QuestLog_Update();
	WatchFrame_Update();
end;

local fragcount = 0;
MQT.Refresh = function(self) 
	MQT.OnToggle(self)
	fragcount = fragcount + 1
	if MQT.db.profile.debug then Print("Aaaaah, refreshing for the " .. fragcount .. "th time!") end --DEBUG
	
	--[[if MQT.db.profile.removecomplete then
		local numQuestWatches = GetNumQuestWatches();
		for i = 1, numQuestWatches do
			local questTitle, _, _, _, _, _, isComplete = GetQuestLogTitle(i);

			if isComplete and MQT.db.profile.removecomplete then
				RemoveQuestWatch(i)
				return
			end			
		end
	end]]
end

function MQT.OnEnable(self)
	self.hooks:enable();
	self:OnToggle();
	Print(L["Moncaí's Quest Tracker enabled"]);
end;

function MQT.OnDisable(self)
	self.hooks:disable();
	self:OnToggle();
	Print(L["Moncaí's Quest Tracker disabled"]);
end;

function MQT.Toggle(self)
	if self:IsEnabled() then
		self:Disable();
	else
		self:Enable();
	end;
end;

function MQT.FitLine(line, maxwidth)
	local text, curlen;
	text = line:GetText();
	curlen = line:GetStringWidth();
	if (curlen > maxwidth) then
		local firstpart, lastword, suc = text, "", false;

		for w,x in text:gmatch(L["(.*)(: ?.*)"]) do
			firstpart, lastword = w, x;
		end
		local lalen = 4;
		for word in firstpart:gmatch(L["([^ ]+)$"]) do
			lalen = word:len();
		end
		lastword = firstpart:sub(firstpart:len() - (lalen-1)) .. lastword; 
		firstpart = firstpart:sub(1, firstpart:len() - (lalen+1)); -- eat a space
		--Print(string.format("We have `%s` `%s` now split up.",firstpart, lastword));

		line:SetText(firstpart .. "..." .. lastword);
		curlen = line:GetStringWidth();
		local fl = firstpart:len() -1;
		while (curlen > maxwidth and fl > 0) do
			firstpart = firstpart:sub(1, fl);
			fl = fl-1;
			line:SetText(firstpart .. "..." .. lastword);
			curlen = line:GetStringWidth();
		end
		if (fl <= 0) then
			line:SetText("< " .. lastword);			
		end
	end
end

function MQT.UpdateWatches()
	MQTW.watches = 0;
	MQT.EnumWatches();
end

--[[ 	MQT.EnumWatches()
	adds the rest of the quests
]]
function MQT.EnumWatches(done)
	local skip = done or {};
	
	local numEntries, numQuests = GetNumQuestLogEntries();
	local questTitle, _, isHeader;
	local W = MQTW;
	
	for i=1, numEntries + numQuests, 1 do
		questTitle, _, _, _, isHeader = GetQuestLogTitle(i);
		if not questTitle then break; end
		
		if (MQTC.QList[questTitle] == true) and (not isHeader) and (not skip[i]) then
			W.watches = W.watches + 1;
			W.hash[W.watches] = i;
		end
	end
	
	--[[ No longer available since 3.2.0
	if W.watches == 0 then
		QuestLogTrackTracking:SetVertexColor(1.0, 0, 0);
	else
		QuestLogTrackTracking:SetVertexColor(0, 1.0, 0);
	end;]]
end
-------------------------------------------------------------------------------
--
-- Builtins override
--
-------------------------------------------------------------------------------

MQT.builtins = {};
local fun = MQT.builtins;

function fun.IsQuestWatched(questID) 
	local questTitle, level, questTag, suggestedGroup, isHeader = GetQuestLogTitle(questID);
	return ( (MQTC.QList[questTitle] == true) and (not isHeader) );
end


function fun.AddQuestWatch(questID, timer, hey)
	local questTitle, _, _, _, _, _, isComplete = GetQuestLogTitle(questID);
	
	--Print(string.format("Really adding %d:%s now", questID, questTitle));
	
	-- timer is set only when invoked from auto_watch thingy
	if (hey and not MQTC.QList[questTitle]) then
		if MQT.db.profile.verbosechat then Print(L["Hey, you're working on %s, I'll add it!"]:format(questTitle)) end
	end

	MQTC.QList[questTitle] = true;
	
	MQT.UpdateWatches();
end

function fun.RemoveQuestWatch(questID)
	local questTitle = GetQuestLogTitle(questID);
	MQTC.QList[questTitle] = nil;

	MQT.UpdateWatches();
end

function fun.GetQuestIndexForWatch(questID)
	return MQTW.hash[questID];
end

function fun.GetNumQuestWatches() 
	MQT.UpdateWatches();
	return MQTW.watches;
end

-------------------------------------------------------------------------------
-- test
-------------------------------------------------------------------------------

function Test() 
	local i=0;
	while (GetQuestLogTitle(i+1) ~= nil) do
		i = i + 1;
		local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(i);
		if (not isHeader) then
			DEFAULT_CHAT_FRAME:AddMessage(questTitle .. " [" .. level .. "]");
		else
			DEFAULT_CHAT_FRAME:AddMessage("## " .. questTitle .. " ##");
		end
	end
	
	local numEntries, numQuests = GetNumQuestLogEntries()
	DEFAULT_CHAT_FRAME:AddMessage(numEntries .. " entries containing " .. numQuests .. " quests in your quest log.");
end
