local ML = LibStub("MoncaiLib");
local MQT = LibStub("AceAddon-3.0"):GetAddon("MoncaiQuestTracker");

local WF = MQT:NewModule("MQT_WatchFrame");
WF.fun = {};
local fun = WF.fun;

--------------------------------------------------------------------------------------------------------
--
-- Overridden functions
--
--------------------------------------------------------------------------------------------------------

-- Some useless local replication here
local watchButtonIndex = 1;
local function WatchFrame_GetLinkButton ()
	local button = WATCHFRAME_LINKBUTTONS[watchButtonIndex]
	if ( not button ) then
		WATCHFRAME_LINKBUTTONS[watchButtonIndex] = WatchFrame.buttonCache:GetFrame();
		button = WATCHFRAME_LINKBUTTONS[watchButtonIndex];
	end

	watchButtonIndex = watchButtonIndex + 1;
	return button;
end

local function WatchFrame_ResetLinkButtons ()
	watchButtonIndex = 1;
end

local function WatchFrame_ReleaseUnusedLinkButtons ()
	local watchButton
	for i = watchButtonIndex, #WATCHFRAME_LINKBUTTONS do
		watchButton = WATCHFRAME_LINKBUTTONS[i];
		if watchButton then
			watchButton.type = nil
			watchButton.index = nil;
			watchButton:Hide();
			watchButton.frameCache:ReleaseFrame(watchButton);
		end
		WATCHFRAME_LINKBUTTONS[i] = nil;
	end
end

local achievementLineIndex = 1;
local function WatchFrame_GetAchievementLine ()
	local line = WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex];
	if ( not line ) then
		WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex] = WatchFrame.lineCache:GetFrame();
		line = WATCHFRAME_ACHIEVEMENTLINES[achievementLineIndex];
	end

	line:Reset();
	achievementLineIndex = achievementLineIndex + 1;
	return line;
end

local function WatchFrame_ResetAchievementLines ()
	achievementLineIndex = 1;
end

local function WatchFrame_ReleaseUnusedAchievementLines ()
	local line
	for i = achievementLineIndex, #WATCHFRAME_ACHIEVEMENTLINES do
		line = WATCHFRAME_ACHIEVEMENTLINES[i];
		line:Hide();
		line.frameCache:ReleaseFrame(line);
		WATCHFRAME_ACHIEVEMENTLINES[i] = nil;
	end
end

local questLineIndex = 1;
local function WatchFrame_GetQuestLine ()
	local line = WATCHFRAME_QUESTLINES[questLineIndex];
	if ( not line ) then
		WATCHFRAME_QUESTLINES[questLineIndex] = WatchFrame.lineCache:GetFrame();
		line = WATCHFRAME_QUESTLINES[questLineIndex];
	end

	line:Reset();
	questLineIndex = questLineIndex + 1;
	
	if MQTC.fontheight then -- MOD
		line.text:SetFontObject(nil);
		line.text:SetFont(STANDARD_TEXT_FONT, MQTC.fontheight);
		local _, fontHeight = line.text:GetFont();
		WATCHFRAMELINES_FONTHEIGHT = fontHeight;
		WATCHFRAMELINES_LINEHEIGHT = fontHeight + 2;
		Printf("We have %d, %d", WATCHFRAMELINES_FONTHEIGHT, WATCHFRAMELINES_LINEHEIGHT)
		WATCHFRAMELINES_FONTSPACING = (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTHEIGHT) / 2
		WATCHFRAME_QUEST_OFFSET = 2;
	end --EMOD
	return line;
end

local function WatchFrame_ResetQuestLines ()
	questLineIndex = 1;
end

local function WatchFrame_ReleaseUnusedQuestLines ()
	local line
	for i = questLineIndex, #WATCHFRAME_QUESTLINES do
		line = WATCHFRAME_QUESTLINES[i];
		line:Hide();
		line.frameCache:ReleaseFrame(line);
		WATCHFRAME_QUESTLINES[i] = nil;
	end
end

-- WatchFrame_GetHeightNeededForQuest
-- from: FrameXML/WatchFrame.lua
-- why: Include quests without objectives
-- changed: track quests without objectives
function fun.WatchFrame_GetHeightNeededForQuest (questIndex)
	local numObjectives = GetNumQuestLeaderBoards(questIndex);
	
	if ( numObjectives == 0 ) then
		return 0;
	else		
		local height = (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) * (numObjectives + 1); -- +1 for the title line
		if ( GetQuestLogSpecialItemInfo(questIndex) ) then
			height = max(height, WATCHFRAME_QUEST_WITH_ITEM_HEIGHT);
		end
		
		local numWatches = GetNumQuestWatches();
		if ( numWatches > 1 ) then
			height = height + WATCHFRAME_QUEST_OFFSET;
		end
		
		return height;
	end
end

function fun.WatchFrame_GetRemainingSpace ()	
	local watchFrame = WatchFrame;
	if ( WATCHFRAME_NUM_TIMERS == 0 ) then
		return 300+watchFrame:GetTop() - watchFrame:GetBottom() - WATCHFRAMELINES_YOFFSET - WATCHFRAME_QUESTTIMER_HEIGHT - math.abs(watchFrame.nextOffset);
	end
	
	return 300+(watchFrame:GetTop() - watchFrame:GetBottom() - WATCHFRAMELINES_YOFFSET) - math.abs(watchFrame.nextOffset);
end

-- WatchFrame_GetHeightNeededForQuest
-- from: FrameXML/WatchFrame.lua
-- why: Include quests without objectives
-- changed: track quests without objectives
function fun.WatchFrame_GetHeightNeededForQuest (questIndex)
	local numObjectives = GetNumQuestLeaderBoards(questIndex);
	
	if ( numObjectives == 0 ) then
		return 0;
	else		
		local height = (WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) * (numObjectives + 1); -- +1 for the title line
		if ( GetQuestLogSpecialItemInfo(questIndex) ) then
			height = max(height, WATCHFRAME_QUEST_WITH_ITEM_HEIGHT);
		end
		
		local numWatches = GetNumQuestWatches();
		if ( numWatches > 1 ) then
			height = height + WATCHFRAME_QUEST_OFFSET;
		end
		
		return height;
	end
end

function fun.WatchFrame_Update (self)
	self = self or WatchFrame; -- Speeds things up if we pass in this reference when we can conveniently.
	-- Display things in this order: quest timers, achievements, quests, addon subscriptions.
	if ( self.updating ) then
		return;
	end
	
	self.updating = true;
	
	local pixelsUsed = 0;
	local totalOffset = WATCHFRAME_INITIAL_OFFSET;
	local lineFrame = WatchFrameLines;
	local maxHeight = (WatchFrame:GetTop() - WatchFrame:GetBottom()); -- Can't use lineFrame:GetHeight() because it could be an invalid rectangle (width of 0)
	
	local maxFrameWidth;
	if ( self.simpleMode ) then
		maxFrameWidth = 1024; -- Effectively meaningless number -- FUTURE MOD
	else
		maxFrameWidth = self:GetWidth() - WATCHFRAMELINES_XOFFSET;
	end
	
	local maxWidth = 0;
	local maxLineWidth;
	
	WatchFrame_ResetLinkButtons();
	
	for i = 1, #WATCHFRAME_OBJECTIVEHANDLERS do
		pixelsUsed, maxLineWidth = WATCHFRAME_OBJECTIVEHANDLERS[i](lineFrame, totalOffset, maxHeight, maxFrameWidth);
		maxWidth = max(maxLineWidth, maxWidth);
		
		if ( pixelsUsed > 0 ) then
			totalOffset = totalOffset - WATCHFRAME_TYPE_OFFSET - pixelsUsed;
		end
	end
	
	WatchFrame_ReleaseUnusedLinkButtons();
	
	if ( self.simpleMode or not self.keepLeft ) then -- Keep right
		if ( not lineFrame.keepRight ) then
			lineFrame.keepRight = true;
			lineFrame.keepLeft = nil;
			lineFrame:ClearAllPoints();
			lineFrame:SetPoint("TOPRIGHT", self, "TOPRIGHT", -12, -30);
			lineFrame:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -12, 12);
		end
		lineFrame:SetWidth(min(maxWidth, maxFrameWidth));
	elseif ( self.keepLeft ) then
		if ( not lineFrame.keepLeft ) then
			lineFrame.keepRight = nil;
			lineFrame.keepLeft = true;
			lineFrame:ClearAllPoints();
			lineFrame:SetPoint("TOPLEFT", self, "TOPLEFT", 14, -30);
			lineFrame:SetPoint("BOTTOMRIGHT", self, "BOTTOMRIGHT", -12, 12);
		end
	end
	
	if ( WATCHFRAME_NUM_TIMERS == 0 ) then
		self:SetMinResize(WATCHFRAME_MINIMUMWIDTH, max(WATCHFRAME_MINIMUMHEIGHT, -totalOffset + WATCHFRAME_QUESTTIMER_HEIGHT + WATCHFRAMELINES_YOFFSET))
	else
		self:SetMinResize(WATCHFRAME_MINIMUMWIDTH, max(WATCHFRAME_MINIMUMHEIGHT, -totalOffset + WATCHFRAMELINES_YOFFSET));
	end
	self.updating = nil;
	self.nextOffset = totalOffset;
end

function fun.WatchFrame_DisplayTrackedAchievements (lineFrame, initialOffset, maxHeight, frameWidth, ...)
	local _; -- Doing this here thanks to IBLJerry!
	local self = WatchFrame;
	local numTrackedAchievements = select("#", ...);
	
	WatchFrame_ResetAchievementLines();
	
	local lineCache = WatchFrame.lineCache;
	local maxWidth = 0;
	local heightUsed = 0;
	local heightNeeded = 0;
	
	local achievementID;
	local achievementName, completed, description, icon;
	
	local line;
	local achievementTitle;
	local previousLine;
	local nextXOffset = 0;
	local linkButton;
		
	local numCriteria, criteriaDisplayed;
	local criteriaString, criteriaType, criteriaCompleted, quantity, totalQuantity, name, flags, assetID, quantityString, criteriaID, achievementCategory;

	local displayOnlyArena = (not WatchFrame:IsUserPlaced()) and ArenaEnemyFrames and ArenaEnemyFrames:IsShown();
	for i = 1, numTrackedAchievements do
		achievementID = select(i, ...);
		achievementCategory = GetAchievementCategory(achievementID);
		if ( (not displayOnlyArena) or achievementCategory == WATCHFRAME_ACHIEVEMENT_ARENA_CATEGORY ) then
			_, achievementName, _, completed, _, _, _, description, _, icon = GetAchievementInfo(achievementID);
			
			local heightNeeded = WatchFrame_GetHeightNeededForAchievement(achievementID);
			if ( heightNeeded > maxHeight + (initialOffset - heightUsed) ) then
				return heightUsed, maxWidth; -- We ran out of space to draw achievements, stop.
			else
				heightUsed = heightUsed + heightNeeded;
			end
			
			line = WatchFrame_GetAchievementLine();
			achievementTitle = line;
			line.text:SetText(achievementName);
			if ( completed ) then
				line.text:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);
			else
				line.text:SetTextColor(0.75, 0.61, 0);
			end
			line:Show()
			local lineWidth = line.text:GetStringWidth() + WATCHFRAME_ICONXOFFSET
			maxWidth = max(maxWidth, lineWidth)
			if ( previousLine ) then -- If this isn't the first displayed title, our position is relative to the last displayed line.
				local yOffset = 0;
				if ( previousLine.statusBar:IsShown() ) then
					yOffset = -5
				end
				
				line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET);
				line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset - WATCHFRAME_QUEST_OFFSET);
			else
				line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset);
				line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset);
			end
			if ( not self.disableButtons ) then
				linkButton = WatchFrame_GetLinkButton();
				linkButton:SetPoint("TOPLEFT", line.text);
				linkButton:SetPoint("BOTTOMLEFT", line.text);
				linkButton:SetWidth(lineWidth + WATCHFRAME_ICONXOFFSET);
				linkButton.type = "ACHIEVEMENT"
				linkButton.index = achievementID;
				linkButton:Show();
			end
			nextXOffset = 0;
			previousLine = line;
			numCriteria = GetAchievementNumCriteria(achievementID);
			if ( numCriteria > 0 ) then
				criteriaDisplayed = 0;
				for j = 1, numCriteria do
					criteriaString, criteriaType, criteriaCompleted, quantity, totalQuantity, name, flags, assetID, quantityString, criteriaID = GetAchievementCriteriaInfo(achievementID, j);
					if ( criteriaCompleted or ( criteriaDisplayed > WATCHFRAME_CRITERIA_PER_ACHIEVEMENT and not criteriaCompleted ) ) then
						-- Don't do anything
					elseif ( criteriaDisplayed == WATCHFRAME_CRITERIA_PER_ACHIEVEMENT ) then
						-- We ran out of space to display incomplete criteria >_<
						line = WatchFrame_GetAchievementLine();
						line.text:SetText(" - ");
						local dashWidth = line.text:GetStringWidth();
						nextXOffset = nextXOffset + dashWidth -- Offset as if there were a dash
						line.text:SetText("...");
						line.text:SetTextColor(0.8, 0.8, 0.8);
						line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, 3);
						line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, 3);
						line:Show();
						maxWidth = max(maxWidth, line.text:GetStringWidth()); -- I can't imagine this happening anytime soon really
						criteriaDisplayed = criteriaDisplayed + 1;
						previousLine = line;
						nextXOffset = -dashWidth;
					else
						if ( WATCHFRAME_TIMEDCRITERIA[criteriaID] ) then
							local timedCriteria = WATCHFRAME_TIMEDCRITERIA[criteriaID]
							line = WatchFrame_GetAchievementLine();
							line.criteriaID = criteriaID;
							line.duration = timedCriteria.duration;
							line.startTime = timedCriteria.startTime;
							
							local yOffset = WATCHFRAMELINES_FONTSPACING;
							if ( previousLine.statusBar:IsShown() ) then
								yOffset = yOffset - 5;
							end
							line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, yOffset);
							line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset);
							line:Show()
							criteriaDisplayed = criteriaDisplayed + 1;
							previousLine = line;
							nextXOffset = 0;
							WatchFrameLines_AddUpdateFunction(WatchFrame_UpdateTimedAchievements);
						end
						if ( bit.band(flags, ACHIEVEMENT_CRITERIA_PROGRESS_BAR) == ACHIEVEMENT_CRITERIA_PROGRESS_BAR ) then
							line = WatchFrame_GetAchievementLine();
							line.statusBar:Show();
							line.statusBar:GetStatusBarTexture():SetVertexColor(0, 0.6, 0, 1);
							line.statusBar:SetMinMaxValues(0, totalQuantity);
							line.statusBar:SetValue(quantity);
							line.statusBar.text:SetText(quantityString);
							
							local yOffset = -5;
							if ( previousLine.statusBar:IsShown() ) then
								yOffset = -10;
							end
							
							line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, yOffset);
							line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset);					
							line:Show();
							maxWidth = max(maxWidth, 200);
							criteriaDisplayed = criteriaDisplayed + 2;
							nextXOffset = 0;
							previousLine = line;
						else
							line = WatchFrame_GetAchievementLine();
							line.text:SetText(" - " .. criteriaString);
							line.text:SetTextColor(0.8, 0.8, 0.8);
							local yOffset = WATCHFRAMELINES_FONTSPACING;
							if ( previousLine.statusBar:IsShown() ) then
								yOffset = yOffset - 5;
							end
							
							line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, yOffset);
							line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset);
							line:Show();
							maxWidth = max(maxWidth, line.text:GetStringWidth());
							criteriaDisplayed = criteriaDisplayed + 1;
							nextXOffset = 0;
							previousLine = line;
						end
					end
				end
			else
				local dash = WatchFrame_GetAchievementLine();
				dash.text:SetText(" - ");
				dash.text:SetTextColor(0.8, 0.8, 0.8);
				local yOffset = WATCHFRAMELINES_FONTSPACING;
				if ( previousLine.statusBar:IsShown() ) then
					yOffset = yOffset - 5;
				end
				
				local dashWidth = dash.text:GetStringWidth();
				
				dash:SetWidth(dashWidth);
				dash:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset);
				dash:Show();
				
				line = WatchFrame_GetAchievementLine();
				line.text:SetText(description);
				line.text:SetTextColor(0.8, 0.8, 0.8);
				
				local stringWidth = line.text:GetStringWidth();
				local desiredWidth = math.ceil(stringWidth + dashWidth); -- This is how long we want the line to be with no wrapping
				
				if ( desiredWidth > maxWidth ) then
					maxWidth = min(desiredWidth, MAX_LONG_CRITERIA_WIDTH);
				end
				
				local linesNeeded = math.ceil(stringWidth / (maxWidth - dashWidth));
				
				line:SetPoint("TOPLEFT", dash, "TOPRIGHT");
				line:SetPoint("TOPRIGHT", previousLine, "TOPRIGHT", 0, yOffset);
				line:SetHeight((WATCHFRAME_LINEHEIGHT - WATCHFRAMELINES_FONTSPACING) + (WATCHFRAMELINES_FONTHEIGHT * (linesNeeded - 1)));
				line:Show();
				
				nextXOffset = -dashWidth;
				previousLine = line;	
				
				for criteriaID, timedCriteria in next, WATCHFRAME_TIMEDCRITERIA do
					if ( timedCriteria.achievementID == achievementID ) then
						line = WatchFrame_GetAchievementLine();
						line.criteriaID = criteriaID;
						line.duration = timedCriteria.duration;
						line.startTime = timedCriteria.startTime;
						
						local yOffset = WATCHFRAMELINES_FONTSPACING;
						if ( previousLine.statusBar:IsShown() ) then
							yOffset = yOffset - 5;
						end
						line:SetPoint("TOPRIGHT", previousLine, "BOTTOMRIGHT", 0, yOffset);
						line:SetPoint("TOPLEFT", previousLine, "BOTTOMLEFT", nextXOffset, yOffset);
						line:Show()
						previousLine = line;
						nextXOffset = 0;
						WatchFrameLines_AddUpdateFunction(WatchFrame_UpdateTimedAchievements);
					end
				end					
			end
		end
	end

	WatchFrame_ReleaseUnusedAchievementLines();

	return heightUsed, maxWidth;
end


-- WatchFrame_DisplayTrackedQuests
-- from: FrameXML/WatchFrame.lua
-- why: Alter appearance
-- changed: some things here and there
function fun.WatchFrame_DisplayTrackedQuests (lineFrame, initialOffset, maxHeight, frameWidth)
	local _, bs, mytitle, color; --MOD
	local self = WatchFrame;
	local text, finished;
	local questTitle;
	local watchItemIndex = 0;
	local questIndex;
	local objectivesCompleted;

	WatchFrame_ResetQuestLines();
	
	local line;
	local lastLine;
	local HIGHLIGHT_FONT_COLOR = HIGHLIGHT_FONT_COLOR;
	local NORMAL_FONT_COLOR = NORMAL_FONT_COLOR;
	local linkButton;
	
	local heightNeeded = 0;
	local heightUsed = 0;
	local maxWidth = 0;
	
	local iconHeightLeft = 0; local bl = MQT.BuddyList; --MOD

	local numQuestWatches = GetNumQuestWatches();
	local numObjectives;	
	local title, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID;
	for i = 1, numQuestWatches do
		local questWidth = 0; --MOD
		questIndex = GetQuestIndexForWatch(i);
		title, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily, questID = GetQuestLogTitle(questIndex);
		if ( questIndex ) then
			local heightNeeded = WatchFrame_GetHeightNeededForQuest(questIndex);
			if ( heightNeeded > maxHeight + (initialOffset - heightUsed) ) then
				-- We ran out of space to draw quests, stop.
				break;
			else
				heightUsed = heightUsed + heightNeeded;
			end

			line = WatchFrame_GetQuestLine();
			questTitle = line;
			
			-- BEGIN MOD
			if (not isComplete) then 
				mytitle = title 
			else 
				if (isComplete < 0) then 
					mytitle = title .. " (" .. FAILED .. ")";
				else
					mytitle = title .. " (" .. COMPLETE .. ")";
				end
			end
			
			if (MQT.db.profile.showlevel) then mytitle = string.format(MQT.LEVEL_FORMAT, level, mytitle); end
			
			line.text:SetText(mytitle);
			
			-- Color title
			if (MQT.db.profile.colortitle) then
				if (isDaily) then
					line.text:SetTextColor(0.3, 0.5, 0.9);
				else
					color = GetQuestDifficultyColor(level);
					line.text:SetTextColor(color.r, color.g, color.b);					
				end
			else
				
			end
			
			-- END MOD
			line:Show();
			if ( not lastLine ) then -- First line
				line:SetPoint("TOPRIGHT", lineFrame, "TOPRIGHT", 0, initialOffset);
				line:SetPoint("TOPLEFT", lineFrame, "TOPLEFT", 0, initialOffset);
			else
				local yOffset = 0;
				if ( iconHeightLeft > 0 ) then
					yOffset = -iconHeightLeft;
				end
				
				line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, yOffset - WATCHFRAME_QUEST_OFFSET);
				line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, yOffset - WATCHFRAME_QUEST_OFFSET);
			end
			iconHeightLeft = 0;
			-- original location of linkbutton code
			local stringWidth = line.text:GetStringWidth();
			
			if ( SHOW_QUEST_OBJECTIVES_ON_MAP == "1") then
				line.mapButton:Show();
				line.mapButton.questID = questID;
				line.text:ClearAllPoints();
				line.text:SetPoint("TOPLEFT");
				local lineWidth = line:GetWidth();
				if ( line.text:GetWidth() > lineWidth ) then
					line.text:SetWidth(lineWidth);
				end
			end

			local objectiveWidth = frameWidth;			
			local link, item, charges = GetQuestLogSpecialItemInfo(questIndex);
			local questWidth = 0;
			local itemButton;
			if ( item ) then
				watchItemIndex = watchItemIndex + 1;
				itemButton = _G["WatchFrameItem"..watchItemIndex];
				if ( not itemButton ) then
					WATCHFRAME_NUM_ITEMS = watchItemIndex;
					itemButton = CreateFrame("BUTTON", "WatchFrameItem" .. watchItemIndex, lineFrame, "WatchFrameItemButtonTemplate");
				end
				itemButton:Show();
				itemButton:ClearAllPoints();
				itemButton:SetID(questIndex);
				SetItemButtonTexture(itemButton, item);
				SetItemButtonCount(itemButton, charges);
				itemButton.charges = charges;
				WatchFrameItem_UpdateCooldown(itemButton);
				itemButton.rangeTimer = -1;
				if ( SHOW_QUEST_OBJECTIVES_ON_MAP ~= "1" ) then
					line.text:SetPoint("RIGHT", itemButton, "LEFT", -4, 0);
				end
				--iconHeightLeft = WATCHFRAME_QUEST_WITH_ITEM_HEIGHT - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING; -- We've already displayed a line for this
				iconHeightLeft = WATCHFRAME_QUEST_WITH_ITEM_HEIGHT - WATCHFRAME_LINEHEIGHT; --MOD
				itemButton.maxStringWidth = stringWidth;
				objectiveWidth = objectiveWidth - (WATCHFRAME_ITEM_WIDTH + 4); --MOD				
			end
			questWidth = max(stringWidth, questWidth);
			
			MQT.FitLine(line.text, objectiveWidth); -- MOD fit questtitle
			
			-- relocated linkbutton code to take linelength after shortening
			-- fixes overlap with itembutton
			stringWidth = line.text:GetStringWidth();
			if ( not self.disableButtons ) then
				linkButton = WatchFrame_GetLinkButton();
				linkButton:SetPoint("TOPLEFT", line);
				linkButton:SetPoint("BOTTOMLEFT", line);
				linkButton:SetPoint("RIGHT", line.text);
				linkButton.type = "QUEST"
				linkButton.index = i; -- We want the Watch index, we'll get the quest index later with GetQuestIndexForWatch(i);
				linkButton:Show();
			end
			
			lastLine = line;
			objectivesCompleted = 0;
			numObjectives = GetNumQuestLeaderBoards(questIndex);
			for j = 1, numObjectives do repeat
				text, _, finished = GetQuestLogLeaderBoard(j, questIndex);
								
				bs = bl:getBuddyWatch(title, j) -- MODIFIED
				if bs ~= "" then
					local w,x,y = string.gmatch(text, MQT.MATCH3)();
					if w and x and y then
						text = string.format("%s: %s%s/%d", w,x,bs,y)
					end
				end -- END MOD
				
				if finished and MQT.db.profile.hidecompletedobjectives and bs == "" then
					objectivesCompleted = objectivesCompleted + 1;
					break -- this is actually a continue working in concerto with a repeat until true
				end
				
				line = WatchFrame_GetQuestLine();
				
				line.text:SetText(" - " .. text);
				MQT.FitLine(line.text, objectiveWidth); --MOD
				if ( finished ) then
					line.text:SetTextColor(HIGHLIGHT_FONT_COLOR.r, HIGHLIGHT_FONT_COLOR.g, HIGHLIGHT_FONT_COLOR.b);
					objectivesCompleted = objectivesCompleted + 1;
				else
					line.text:SetTextColor(0.8, 0.8, 0.8);
				end
				line:SetPoint("TOPRIGHT", lastLine, "BOTTOMRIGHT", 0, WATCHFRAMELINES_FONTSPACING);
				line:SetPoint("TOPLEFT", lastLine, "BOTTOMLEFT", 0, WATCHFRAMELINES_FONTSPACING);
				line:Show();
				stringWidth = line.text:GetStringWidth();
				if ( iconHeightLeft > 0 ) then
					if ( SHOW_QUEST_OBJECTIVES_ON_MAP == "1" ) then
						line.text:SetPoint("RIGHT", itemButton, "LEFT", -WATCHFRAME_MAP_BUTTON_WIDTH, 0);
					else
						line.text:SetPoint("RIGHT", itemButton, "LEFT", -4, 0);
					end
					itemButton.maxStringWidth = max(stringWidth, itemButton.maxStringWidth)
				end
				questWidth = max(stringWidth, questWidth);
				lastLine = line;
				--iconHeightLeft = iconHeightLeft - WATCHFRAMELINES_FONTHEIGHT - WATCHFRAMELINES_FONTSPACING;
				iconHeightLeft = iconHeightLeft - WATCHFRAME_LINEHEIGHT;				
			until true; end

			-- NOTE: we're missing something to display required money for a quest...that should probably be added at some point

			if not MQT.db.profile.colortitle then -- MOD
				if ( isComplete ) then
					questTitle.text:SetTextColor(NORMAL_FONT_COLOR.r, NORMAL_FONT_COLOR.g, NORMAL_FONT_COLOR.b);
				else
					questTitle.text:SetTextColor(0.75, 0.61, 0);
				end
			end -- MOD
				
			local itemButtonXOffset = 8;
			if ( SHOW_QUEST_OBJECTIVES_ON_MAP == "1") then
				questWidth = questWidth + WATCHFRAME_MAP_BUTTON_WIDTH;
				itemButtonXOffset = itemButtonXOffset + WATCHFRAME_MAP_BUTTON_WIDTH;
			end
			if ( itemButton ) then
				questWidth = questWidth + WATCHFRAME_ITEM_WIDTH;
				itemButtonXOffset = min(frameWidth - WATCHFRAME_ITEM_WIDTH, itemButton.maxStringWidth + itemButtonXOffset);
				itemButton:SetPoint("TOPLEFT", questTitle, "TOPLEFT", itemButtonXOffset, -WATCHFRAMELINES_FONTSPACING);
				itemButton:Show();
				if ( SHOW_QUEST_OBJECTIVES_ON_MAP == "1" and (questTitle.text:GetWidth() + WATCHFRAME_MAP_BUTTON_WIDTH) > itemButtonXOffset ) then
					questTitle.text:SetPoint("RIGHT", itemButton, "LEFT", -WATCHFRAME_MAP_BUTTON_WIDTH, 0);
				end
			end
			maxWidth = max(questWidth, maxWidth);
		end
	end

	for i = watchItemIndex + 1, WATCHFRAME_NUM_ITEMS do
		_G["WatchFrameItem" .. i]:Hide();
	end

	WatchFrame_ReleaseUnusedQuestLines();

	return heightUsed, maxWidth;
end

--------------------------------------------------------------------------------------------------------
--
-- Initialize
--
--------------------------------------------------------------------------------------------------------

function WF.OnInitialize(self)	
	local hooks = ML.api.hooking:new(self);
	self.hooks = hooks;
	hooks:HookGlobals(self.fun);
end

function WF.OnEnable(self)
	WatchFrame_RemoveObjectiveHandler(WatchFrame_DisplayTrackedQuests)
	self.hooks:enable();
	WatchFrame_AddObjectiveHandler(WatchFrame_DisplayTrackedQuests);	
end

function WF.OnDisable(self)
	WatchFrame_RemoveObjectiveHandler(WatchFrame_DisplayTrackedQuests)
	self.hooks:disable();
	WatchFrame_AddObjectiveHandler(WatchFrame_DisplayTrackedQuests);
end
